(* Quickcheck utils for subrange_by_rank, used in test and benchmarks *)

open Core
open Quickcheck
open Import
module Key = Int
module Value = Int

type map = Value.t Key.Map.t [@@deriving sexp_of, compare]

type map_op =
  [ `Add_nearby of Key.t * Value.t
  | `Remove of Key.t
  | `Update of Key.t * Value.t
  | `No_change
  ]
[@@deriving sexp_of]

type range = int Maybe_bound.t * int Maybe_bound.t [@@deriving sexp_of]

val map_with_length_gen : ?key_range:int * int -> int -> map Generator.t
val map_gen : map Generator.t
val map_op_gen : ?key_range:int * int -> ?none_ratio:float -> unit -> map_op Generator.t

type range_op =
  [ `Move_start of int
  | `Move_end of int
  | `Next_page
  | `Prev_page
  | `Set_range of range (* not generated by default *)
  | `No_change
  ]
[@@deriving sexp_of]

val range_gen : range Generator.t
val range_op_gen : ?none_ratio:float -> unit -> range_op Generator.t
val apply_map_op : map -> map_op -> map
val apply_map_op_incr : map Incr.Var.t -> map_op -> unit
val apply_range_op : range -> range_op -> range
val apply_range_op_incr : range Incr.Var.t -> range_op -> unit
val map_and_range_op_gen : ?key_range:int * int -> unit -> (map_op * range_op) Generator.t
